import typing
from typing import Any

import jax
import jax.numpy as jnp


if getattr(typing, "GENERATING_DOCUMENTATION", False):
    Array = "jax.numpy.ndarray"
else:
    Array = jnp.ndarray

PyTree = Any

TreeDef = type(jax.tree_structure(0))



import functools as ft
from dataclasses import dataclass
from typing import Any

import jax

from .deprecated import deprecated
from .filters import combine, is_array, partition, validate_filters
from .module import Module, static_field


class _Static(Module):
    value: Any = static_field()


@ft.lru_cache(maxsize=4096)
def _filter_jit_cache(f, **jitkwargs):
    @ft.partial(jax.jit, static_argnums=(0, 1, 4), **jitkwargs)
    def f_wrapped(
        static_leaves, static_treedef, dynamic_args, dynamic_kwargs, filter_spec_return
    ):
        static_args, static_kwargs = jax.tree_unflatten(static_treedef, static_leaves)
        args = combine(dynamic_args, static_args)
        kwargs = combine(dynamic_kwargs, static_kwargs)
        out = f(*args, **kwargs)
        dynamic_out, static_out = partition(out, filter_spec_return)
        return dynamic_out, _Static(static_out)

    return f_wrapped


def filter_jit(fun, *, filter_spec=is_array, filter_spec_return=is_array, **jitkwargs):
    """Wraps together [`equinox.partition`][] and `jax.jit`.

    **Arguments:**

    - `fun` is a pure function to JIT compile.
    - `filter_spec` is a PyTree whose structure should be a prefix of the structure of
        the inputs to `fun`. It behaves as the `filter_spec` argument to
        [`equinox.filter`][]. Truthy values will be traced; falsey values will be held
        static.
    - `filter_spec_return` is a PyTree whose structure should be a prefix of the
        structure of the outputs of `fun`. It behaves as the `filter_spec` argument to
        [`equinox.filter`][]. Truthy values should be tracers; falsely values are any
        (non-tracer) auxiliary information to return.
    - `**jitkwargs` are any other keyword arguments to `jax.jit`.

        !!! info

            Specifically, if calling `fun(*args, **kwargs)`, then `filter_spec` must
            have a structure which is a prefix for `(args, kwrgs)`.

    **Returns:**

    The JIT'd version of `fun`.

    !!! info

        A very important special case is to trace all JAX arrays and treat all other
        objects as static.

        This is accomplished with `filter_spec=equinox.is_array`,
        `filter_spec_return=equinox.is_array` -- which are the defaults. (It is
        unusual to need different behaviour to this.)
    """

    if any(
        x in jitkwargs for x in ("static_argnums", "static_argnames", "donate_argnums")
    ):
        raise ValueError(
            "`jitkwargs` cannot contain 'static_argnums', 'static_argnames' or "
            "'donate_argnums'."
        )

    # We choose not to make a distinction between ([arg, ..., arg], kwargs) and ((arg, ..., arg), kwargs)
    if (
        isinstance(filter_spec, tuple)
        and len(filter_spec) == 2
        and isinstance(filter_spec[0], list)
    ):
        filter_spec = (tuple(filter_spec[0]), filter_spec[1])

    @ft.wraps(fun)
    def fun_wrapper(*args, **kwargs):
        (dynamic_args, dynamic_kwargs), (static_args, static_kwargs) = partition(
            (args, kwargs), filter_spec
        )
        static_leaves, static_treedef = jax.tree_flatten((static_args, static_kwargs))
        static_leaves = tuple(static_leaves)
        dynamic_out, static_out = _filter_jit_cache(fun, **jitkwargs)(
            static_leaves,
            static_treedef,
            dynamic_args,
            dynamic_kwargs,
            filter_spec_return,
        )
        return combine(dynamic_out, static_out.value)

    return fun_wrapper


#
# Deprecated
#


@ft.lru_cache(maxsize=4096)
def _jitf_cache(f, args_treedef, **jitkwargs):
    @ft.partial(jax.jit, **jitkwargs)
    def f_wrapped(*args):
        args = jax.tree_unflatten(args_treedef, args)
        return f(*args)

    return f_wrapped


@dataclass(frozen=True)
class _UnPyTreeAble:
    value: Any

    def __bool__(self):
        return False


_marker_sentinel = object()


@deprecated(in_favour_of=filter_jit)
def jitf(
    fun,
    *,
    filter_fn=None,
    filter_tree=None,
    static_argnums=None,
    static_argnames=None,
    donate_argnums=(),
    **jitkwargs
):
    if isinstance(static_argnums, int):
        static_argnums = (static_argnums,)
    if static_argnames is not None:
        raise NotImplementedError(
            "jitf does not yet support `static_argnames`. use static_argnums instead."
        )
    if donate_argnums != ():
        raise NotImplementedError("jitf does not ye support `donate_argnums`.")
    validate_filters("jitf", filter_fn, filter_tree)

    if static_argnums is None:
        len_static_argnums = 0
    else:
        len_static_argnums = len(static_argnums)

    @ft.wraps(fun)
    def f_wrapper(*args, **kwargs):
        if len(kwargs):
            raise NotImplementedError(
                "jitf does not yet support keyword arguments. Use positional arguments instead."
            )

        if filter_tree is not None:
            if len(args) - len_static_argnums == 1:
                new_filter_tree = (filter_tree,)
            else:
                new_filter_tree = tuple(filter_tree)

        # Mark the arguments that have been explicitly declared static via `static_argnums`
        if static_argnums is not None:
            args = list(args)
            for index in static_argnums:
                args[index] = _UnPyTreeAble(args[index])
            if filter_tree is not None:
                new_filter_tree = list(new_filter_tree)
                for index in static_argnums:
                    new_filter_tree.insert(index, _UnPyTreeAble(None))

        # Flatten everything else
        args_flat, args_treedef = jax.tree_flatten(args)
        if filter_tree is not None:
            filter_flat, flat_treedef = jax.tree_flatten(new_filter_tree)
            if flat_treedef != args_treedef:
                raise ValueError(
                    "The tree stucture for the filters and the arguments must be the same."
                )

        # Figure out static argnums with respect to this new flattened structure.
        new_static_argnums = []
        if filter_tree is None:
            # implies filter_fn is not None
            for i, arg in enumerate(args_flat):
                if isinstance(arg, _UnPyTreeAble) or not filter_fn(arg):
                    new_static_argnums.append(i)
        else:
            for i, (arg, filter) in enumerate(zip(args_flat, filter_flat)):
                if not filter:
                    new_static_argnums.append(i)
        new_static_argnums = tuple(new_static_argnums)
        if static_argnums is not None:
            args_flat = [
                arg.value if isinstance(arg, _UnPyTreeAble) else arg
                for arg in args_flat
            ]

        f_jitted = _jitf_cache(
            fun, args_treedef, static_argnums=new_static_argnums, **jitkwargs
        )
        return f_jitted(*args_flat)

    return f_wrapper



import functools as ft
import warnings


def deprecated(*, in_favour_of):
    if not isinstance(in_favour_of, str):
        in_favour_of = in_favour_of.__name__

    def decorator(fn):
        msg = f"{fn.__name__} is deprecated in favour of {in_favour_of}"

        @ft.wraps(fn)
        def wrapper(*args, **kwargs):
            warnings.warn(msg)
            return fn(*args, **kwargs)

        return wrapper

    return decorator



import collections
from itertools import repeat
from typing import Any, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
import jax.random as jrandom
import numpy as np
from jax.lax import conv_general_dilated

from ..custom_types import Array
from ..module import Module, static_field


def _ntuple(n: int) -> callable:
    def parse(x: Any) -> tuple:
        if isinstance(x, collections.abc.Iterable):
            if len(x) == n:
                return tuple(x)
            else:
                raise ValueError(
                    f"Length of {x} (length = {len(x)}) is not equal to {n}"
                )
        return tuple(repeat(x, n))

    return parse


class Conv(Module):
    """General N-dimensional convolution."""

    num_spatial_dims: int = static_field()
    weight: Array
    bias: Optional[Array]
    in_channels: int = static_field()
    out_channels: int = static_field()
    kernel_size: Tuple[int] = static_field()
    stride: Tuple[int] = static_field()
    padding: Tuple[int] = static_field()
    dilation: Tuple[int] = static_field()
    use_bias: bool = static_field()

    def __init__(
        self,
        num_spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Sequence[int]],
        stride: Union[int, Sequence[int]] = 1,
        padding: Union[int, Sequence[int]] = 0,
        dilation: Union[int, Sequence[int]] = 1,
        use_bias: bool = True,
        *,
        key: "jax.random.PRNGKey",
        **kwargs,
    ):
        """**Arguments:**

        - `num_spatial_dims`: The number of spatial dimensions. For example traditional
            convolutions for image processing have this set to `2`.
        - `in_channels`: The number of input channels.
        - `out_channels`: The number of output channels.
        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The amount of padding to apply before and after each spatial
            dimension. The same amount of padding is applied both before and after.
        - `dilation`: The dilation of the convolution.
        - `use_bias`: Whether to add on a bias after the convolution.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)

        !!! info

            All of `kernel_size`, `stride`, `padding`, `dilation` can be either an
            integer or a sequence of integers. If they are a sequence then the sequence
            should be of length equal to `num_spatial_dims`, and specify the value of
            each property down each spatial dimension in turn.. If they are an integer
            then the same kernel size / stride / padding / dilation will be used along
            every spatial dimension.

        """
        super().__init__(**kwargs)
        self.num_spatial_dims = num_spatial_dims
        parse = _ntuple(self.num_spatial_dims)
        wkey, bkey = jrandom.split(key, 2)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = parse(kernel_size)
        self.use_bias = use_bias
        lim = 1 / np.sqrt(self.in_channels * np.prod(self.kernel_size))

        self.weight = jrandom.uniform(
            wkey,
            (
                self.out_channels,
                self.in_channels,
            )
            + self.kernel_size,
            minval=-lim,
            maxval=lim,
        )
        if self.use_bias:
            self.bias = jrandom.uniform(
                bkey,
                (self.out_channels,) + (1,) * self.num_spatial_dims,
                minval=-lim,
                maxval=lim,
            )
        else:
            self.bias = None

        self.stride = parse(stride)
        if isinstance(padding, int):
            self.padding = tuple(
                (padding, padding) for _ in range(self.num_spatial_dims)
            )
        elif isinstance(padding, Sequence) and len(padding) == self.num_spatial_dims:
            self.padding = tuple((p, p) for p in padding)
        else:
            raise ValueError(
                "`padding` must either be an int or tuple of length "
                f"{self.num_spatial_dims}."
            )
        self.dilation = parse(dilation)

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape `(in_channels, dim_1, ..., dim_N)`, where
            `N = num_spatial_dims`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(out_channels, new_dim_1, ..., new_dim_N)`.
        """

        unbatched_rank = self.num_spatial_dims + 1
        if x.ndim != unbatched_rank:
            raise ValueError(
                f"Input to `Conv` needs to have rank {unbatched_rank},",
                f" but input has shape {x.shape}.",
            )
        x = jnp.expand_dims(x, axis=0)
        x = conv_general_dilated(
            lhs=x,
            rhs=self.weight,
            window_strides=self.stride,
            padding=self.padding,
            rhs_dilation=self.dilation,
        )
        if self.use_bias:
            x += jnp.broadcast_to(self.bias, x.shape)
        x = jnp.squeeze(x, axis=0)
        return x


class Conv1d(Conv):
    """As [`equinox.nn.Conv`][] with `num_spatial_dims=1`."""

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        use_bias=True,
        *,
        key,
        **kwargs,
    ):
        super().__init__(
            num_spatial_dims=1,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            use_bias=use_bias,
            key=key,
            **kwargs,
        )


class Conv2d(Conv):
    """As [`equinox.nn.Conv`][] with `num_spatial_dims=2`."""

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=(1, 1),
        padding=(0, 0),
        dilation=(1, 1),
        use_bias=True,
        *,
        key,
        **kwargs,
    ):
        super().__init__(
            num_spatial_dims=2,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            use_bias=use_bias,
            key=key,
            **kwargs,
        )


class Conv3d(Conv):
    """As [`equinox.nn.Conv`][] with `num_spatial_dims=3`."""

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=(1, 1, 1),
        padding=(0, 0, 0),
        dilation=(1, 1, 1),
        use_bias=True,
        *,
        key,
        **kwargs,
    ):
        super().__init__(
            num_spatial_dims=3,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            use_bias=use_bias,
            key=key,
            **kwargs,
        )



from . import nn
from .filters import (
    combine,
    filter,
    is_array,
    is_array_like,
    is_inexact_array,
    is_inexact_array_like,
    merge,
    partition,
    split,
)
from .grad import (
    filter_custom_vjp,
    filter_grad,
    filter_value_and_grad,
    gradf,
    value_and_grad_f,
)
from .jit import filter_jit, jitf
from .module import Module, static_field
from .tree import tree_at, tree_equal
from .update import apply_updates


__version__ = "0.1.4"
